# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.
"""This module is only here to warn users that the `coordinated_workers` have been moved out of cosl."""

raise ImportError(
    "The `coordinated_workers` module has been removed from `cosl`. "
    "Use `https://github.com/canonical/cos-coordinated-workers` instead."
)



# Copyright 2020 Canonical Ltd.
# See LICENSE file for licensing details.

import subprocess
import unittest
import unittest.mock

from cosl import CosTool


class TestTransformPromQL(unittest.TestCase):
    """Test that the cos-tool promql implementation works."""

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    @unittest.mock.patch("subprocess.run")
    def test_returns_original_expression_when_subprocess_call_errors(
        self, mocked_run: unittest.mock.Mock
    ):
        mocked_run.side_effect = subprocess.CalledProcessError(
            returncode=10, cmd="cos-tool", stderr=""
        )

        tool = CosTool(default_query_type="promql")
        output = tool.apply_label_matchers(
            {
                "groups": [
                    {
                        "alert": "CPUOverUse",
                        "expr": "process_cpu_seconds_total > 0.12",
                        "for": "0m",
                        "labels": {
                            "severity": "Low",
                            "juju_model": "None",
                            "juju_model_uuid": "f2c1b2a6-e006-11eb-ba80-0242ac130004",
                            "juju_application": "consumer-tester",
                        },
                        "annotations": {
                            "summary": "Instance {{ $labels.instance }} CPU over use",
                            "description": "{{ $labels.instance }} of job "
                            "{{ $labels.job }} has used too much CPU.",
                        },
                    }
                ]
            },
        )
        self.assertEqual(output["groups"][0]["expr"], "process_cpu_seconds_total > 0.12")

    @unittest.mock.patch("platform.machine", lambda: "invalid")
    def test_uses_original_expression_when_binary_missing(self):
        tool = CosTool(default_query_type="promql")
        output: str = tool.apply_label_matchers(
            {
                "groups": [
                    {
                        "alert": "CPUOverUse",
                        "expr": "process_cpu_seconds_total > 0.12",
                        "for": "0m",
                        "labels": {
                            "severity": "Low",
                            "juju_model": "None",
                            "juju_model_uuid": "f2c1b2a6-e006-11eb-ba80-0242ac130004",
                            "juju_application": "consumer-tester",
                        },
                        "annotations": {
                            "summary": "Instance {{ $labels.instance }} CPU over use",
                            "description": "{{ $labels.instance }} of job "
                            "{{ $labels.job }} has used too much CPU.",
                        },
                    }
                ]
            },
        )
        self.assertEqual(output["groups"][0]["expr"], "process_cpu_seconds_total > 0.12")

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_fetches_the_correct_expression(self):
        tool = CosTool(default_query_type="promql")

        output = tool.inject_label_matchers("up", {"juju_model": "some_juju_model"})
        assert output == 'up{juju_model="some_juju_model"}'

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_handles_comparisons(self):
        tool = CosTool(default_query_type="promql")
        output = tool.inject_label_matchers(
            "up > 1",
            {"juju_model": "some_juju_model"},
        )
        assert output == 'up{juju_model="some_juju_model"} > 1'

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_handles_multiple_labels(self):
        tool = CosTool(default_query_type="promql")
        output = tool.inject_label_matchers(
            "up > 1",
            {
                "juju_model": "some_juju_model",
                "juju_model_uuid": "123ABC",
                "juju_application": "some_application",
                "juju_unit": "some_application/1",
            },
        )
        assert (
            output == 'up{juju_application="some_application",juju_model="some_juju_model"'
            ',juju_model_uuid="123ABC",juju_unit="some_application/1"} > 1'
        )


class TestValidateAlerts(unittest.TestCase):
    """Test that the cos-tool validation works."""

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_returns_errors_on_bad_rule_file(self):
        tool = CosTool(default_query_type="promql")
        valid, errs = tool.validate_alert_rules(
            {
                "groups": [
                    {
                        "alert": "BadSyntax",
                        "expr": "process_cpu_seconds_total{) > 0.12",
                    }
                ]
            },
        )
        self.assertEqual(valid, False)
        self.assertIn("error validating", errs)

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_successfully_validates_good_alert_rules(self):
        tool = CosTool(default_query_type="promql")
        valid, errs = tool.validate_alert_rules(
            {
                "groups": [
                    {
                        "name": "group_name",
                        "rules": [
                            {
                                "alert": "CPUOverUse",
                                "expr": "process_cpu_seconds_total > 0.12",
                                "for": "0m",
                                "labels": {
                                    "severity": "Low",
                                    "juju_model": "None",
                                    "juju_model_uuid": "f2c1b2a6-e006-11eb-ba80-0242ac130004",
                                    "juju_application": "consumer-tester",
                                },
                                "annotations": {
                                    "summary": "Instance {{ $labels.instance }} CPU over use",
                                    "description": "{{ $labels.instance }} of job "
                                    "{{ $labels.job }} has used too much CPU.",
                                },
                            }
                        ],
                    }
                ]
            }
        )
        self.assertEqual(errs, "")
        self.assertEqual(valid, True)



# Copyright 2026 Canonical Ltd.
# See LICENSE file for licensing details.
"""Types used by cos-lib."""

from typing import Any, Dict, List, Literal, Union

from ops.framework import StoredDict, StoredList
from typing_extensions import NotRequired, Required, TypedDict

QueryType = Literal["logql", "promql"]
RuleType = Literal["alert", "record"]


class _RecordingRuleFormat(TypedDict):
    record: Required[str]
    expr: Required[str]
    labels: NotRequired[Dict[str, str]]


class _AlertingRuleFormat(TypedDict):
    alert: Required[str]
    expr: Required[str]
    duration: NotRequired[str]
    keep_firing_for: NotRequired[str]
    labels: NotRequired[Dict[str, str]]
    annotations: NotRequired[Dict[str, str]]


SingleRuleFormat = Union[_AlertingRuleFormat, _RecordingRuleFormat]


class OfficialRuleFileItem(TypedDict):
    """Typing for a single node of the official rule file format."""

    name: str
    rules: List[SingleRuleFormat]


class OfficialRuleFileFormat(TypedDict, total=False):
    """Typing for the official rule file format.

    References:
    - https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/
    - https://prometheus.io/docs/prometheus/latest/configuration/recording_rules/
    """

    groups: List[OfficialRuleFileItem]


def type_convert_stored(
    obj: Union[StoredList, StoredDict, Any],
) -> Union[List[Any], Dict[Any, Any], Any]:
    """Helper for converting Stored[Dict|List|Set] to the objects they pretend to be.

    Ref: https://github.com/canonical/operator/pull/572
    """
    if isinstance(obj, StoredList):
        return list(map(type_convert_stored, obj))
    if isinstance(obj, StoredDict):
        rdict: Dict[Any, Any] = {}
        for k in obj.keys():
            rdict[k] = type_convert_stored(obj[k])
        return rdict
    return obj



import itertools
from typing import Type

import ops

from cosl.reconciler import all_events, reconcilable_events_k8s, reconcilable_events_machine


def _get_inheritance_tree_leaves(cl: Type):
    return list(
        itertools.chain(
            *[
                (
                    [subc]
                    if (subc.__module__.startswith("ops.") and not subc.__subclasses__())
                    else _get_inheritance_tree_leaves(subc)
                )
                for subc in cl.__subclasses__()
            ]
        )
    )


EXCLUDED_EVENTS = {
    ops.CollectMetricsEvent,
}

EXCLUDED_EVENTS_K8S = EXCLUDED_EVENTS.union(
    {
        ops.PebbleCustomNoticeEvent,
        ops.RemoveEvent,
        ops.UpgradeCharmEvent,
    }
)

EXCLUDED_EVENTS_VM = EXCLUDED_EVENTS.union(
    {
        ops.InstallEvent,
        ops.RemoveEvent,
        ops.StartEvent,
        ops.StopEvent,
        ops.UpgradeCharmEvent,
    }
)


def test_correctness():
    """Verify we are surfacing only valid events."""
    all_event_types = set(_get_inheritance_tree_leaves(ops.EventBase))
    assert set(all_events).issubset(all_event_types)


def test_completeness():
    """Verify we are surfacing all events we care about.

    If a new version of ops adds more event types, this will start failing.
    Then we'll have to make a
    choice about whether to put those events in the safe or unsafe bucket.
    """
    all_event_types = set(_get_inheritance_tree_leaves(ops.HookEvent))
    assert set(all_events).union(EXCLUDED_EVENTS) == all_event_types
    assert set(reconcilable_events_k8s).union(EXCLUDED_EVENTS_K8S) == all_event_types
    assert set(reconcilable_events_machine).union(EXCLUDED_EVENTS_VM) == all_event_types


def test_exclusiveness():
    """Verify the safe and unsafe buckets have no intersection."""
    assert set(all_events).intersection(EXCLUDED_EVENTS) == set()
    assert set(reconcilable_events_k8s).intersection(EXCLUDED_EVENTS_K8S) == set()
    assert set(reconcilable_events_machine).intersection(EXCLUDED_EVENTS_VM) == set()



# Copyright 2025 Canonical Ltd.
# See LICENSE file for licensing details.


import pytest

from cosl.time_validation import is_valid_timespec


@pytest.mark.parametrize(
    "given_time,expected_validity",
    [
        ("0", True),
        ("1y", True),
        ("1m", True),
        ("1w", True),
        ("1d", True),
        ("1h", True),
        ("1m", True),
        ("1s", True),
        ("1ms", True),
        ("1sdgs", False),
        ("1w2d", False),
        ("one week", False),
        ("one hour", False),
    ],
)
def test_is_valid_timespec(given_time, expected_validity):
    assert is_valid_timespec(given_time) == expected_validity



# Copyright 2025 Canonical Ltd.
# See LICENSE file for licensing details.
"""Helper function(s) for determining for determining the validity of time options."""

import re

_VALID_TIMESPEC_RE = re.compile(r"^(0|[0-9]+(y|w|d|h|m|s|ms))$")


def is_valid_timespec(timeval: str) -> bool:
    """Returns a boolean based on whether the passed parameter is a valid timespec.

    Upstream ref: https://github.com/prometheus/prometheus/blob/c40e269c3e514953299e9ba1f6265e067ab43e64/cmd/prometheus/main.go#L302

    Args:
        timeval: a string representing a time specification, e.g., "1d", "1w".

    Returns:
        True if time specification is valid and False otherwise.
        The regex in this function returns True when the parameter is 0 or when it uses one of the following units:
        - y for years
        - m for months
        - w for weeks
        - d for days
        - h for hours
        - m for minutes
        - s for seconds
        - ms for milliseconds.
        Otherwise, it returns False.
    """
    return bool(_VALID_TIMESPEC_RE.match(timeval))



from unittest.mock import MagicMock, patch

import ops.framework
import pytest
from ops.testing import Context, Relation, State
from scenario.context import CharmEvents

from cosl.reconciler import all_events, observe_events


def get_observed_events(observe_mock):
    return {call.args[0].event_type for call in observe_mock.call_args_list}


@pytest.fixture
def observe_mock():
    with patch("ops.framework.Framework.observe", MagicMock()) as mm:
        yield mm


@pytest.fixture
def charm(observe_mock):
    ctx = Context(
        ops.CharmBase,
        meta={
            "name": "luca",
            "requires": {"bax": {"interface": "bar"}},
            "containers": {"foo": {}},
            "storage": {"foo": {"type": "bar"}},
        },
        actions={"foo": {}},
    )
    with ctx(ctx.on.update_status(), state=State()) as mgr:
        yield mgr.charm


@pytest.mark.parametrize("event_arg", (True, False))
def test_observe_events(charm, observe_mock, event_arg):
    # GIVEN a regular luca charm
    # WHEN we observe_events with no custom filters
    observe_events(
        charm, events=all_events, handler=(lambda _: None) if event_arg else (lambda: None)
    )
    # THEN events from all categories are observed
    observed_events = get_observed_events(observe_mock)
    assert observed_events == all_events


@pytest.mark.parametrize("event_arg", (True, False))
def test_observe_emission(event_arg):
    # GIVEN a regular luca charm that only observes certain event types
    class LucaCharm(ops.CharmBase):
        def __init__(self, *a, **kw):
            super().__init__(*a, **kw)
            observe_events(
                self,
                events={ops.RelationEvent, ops.SecretEvent, ops.StorageEvent},
                handler=self._event_observer if event_arg else self._reconcile,
            )

        def _event_observer(self, _):  # observe target
            self._reconcile()

        def _reconcile(self):  # observe target
            self.reconcile()

        def reconcile(self):  # mock target (for testing)
            pass

    # WHEN we observe_events
    ctx = Context(
        LucaCharm,
        meta={
            "name": "luca",
            "requires": {"bax": {"interface": "bar"}},
            "containers": {"foo": {}},
        },
        actions={"foo": {}},
    )
    with patch.object(LucaCharm, "reconcile", MagicMock()) as mm:
        # THEN the reconciler does NOT get called on excluded events
        ctx.run(ctx.on.upgrade_charm(), state=State())
        ctx.run(ctx.on.update_status(), state=State())
        ctx.run(ctx.on.install(), state=State())
        ctx.run(ctx.on.stop(), state=State())
        assert not mm.called

    with patch.object(LucaCharm, "reconcile", MagicMock()) as mm:
        relation = Relation("bax")
        ctx.run(ctx.on.relation_changed(relation), state=State(relations={relation}))
        # THEN the reconciler gets called on included events
        assert mm.called


@pytest.mark.parametrize(
    "event, name",
    (
        (CharmEvents.install(), "install"),
        (CharmEvents.upgrade_charm(), "upgradecharm"),
        (CharmEvents.update_status(), "updatestatus"),
    ),
)
def test_observe_groups(event, name):
    # GIVEN a regular luca charm that only observes certain event types
    class LucaCharm(ops.CharmBase):
        def __init__(self, *a, **kw):
            super().__init__(*a, **kw)
            observe_events(
                self,
                events=(ops.InstallEvent,),
                handler=self._install,
            )
            observe_events(
                self,
                events=[ops.UpgradeCharmEvent],
                handler=self._upgradecharm,
            )
            observe_events(
                self,
                events={ops.UpdateStatusEvent},
                handler=self._updatestatus,
            )

        def _updatestatus(self):  # observe target
            self.updatestatus()

        def _upgradecharm(self):  # observe target
            self.upgradecharm()

        def _install(self):  # observe target
            self.install()

        def updatestatus(self):  # mock target (for testing)
            pass

        def upgradecharm(self):  # mock target (for testing)
            pass

        def install(self):  # mock target (for testing)
            pass

    # WHEN we observe_events
    ctx = Context(
        LucaCharm,
        meta={
            "name": "luca",
            "requires": {"bax": {"interface": "bar"}},
            "containers": {"foo": {}},
        },
        actions={"foo": {}},
    )
    mocks = [MagicMock(), MagicMock(), MagicMock()]
    with patch.object(LucaCharm, "updatestatus", mocks[0]):
        with patch.object(LucaCharm, "upgradecharm", mocks[1]):
            with patch.object(LucaCharm, "install", mocks[2]):
                # THEN the right reconciler gets called
                ctx.run(event, state=State())

                expected = {
                    "install": [False, False, True],
                    "upgradecharm": [False, True, False],
                    "updatestatus": [True, False, False],
                }[name]
                calls = [mock.called for mock in mocks]
                assert calls == expected



#!/usr/bin/env python3
# Copyright 2024 Canonical
# See LICENSE file for licensing details.

"""Shared utilities for the inter-coordinator "grafana_datasource_exchange" interface.

See https://github.com/canonical/charm-relation-interfaces/tree/main/interfaces/grafana_datasource_exchange/v0
for the interface specification.
"""

# FIXME: the interfaces import (because it's a git dep perhaps?)
#  can't be type-checked, which breaks everything
# pyright: reportMissingImports=false
# pyright: reportUntypedBaseClass=false
# pyright: reportUnknownLambdaType=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownParameterType=false


import json
import logging
from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
)

import ops
from ops import CharmBase
from pydantic import BaseModel, Field, Json
from typing_extensions import TypedDict

import cosl.interfaces.utils
from cosl.interfaces.utils import DataValidationError

log = logging.getLogger("datasource_exchange")

DS_EXCHANGE_INTERFACE_NAME = "grafana_datasource_exchange"


# FIXME copy-pasta'd from charm-relation-interfaces. Keep in sync!
#  see https://github.com/canonical/charm-relation-interfaces/issues/213
class GrafanaDatasource(BaseModel):
    """GrafanaDatasource model."""

    type: str = Field(
        description="Type of the datasource, typically one of "
        "https://grafana.com/docs/grafana/latest/datasources/#built-in-core-data-sources.",
        examples=["tempo", "loki", "prometheus", "elasticsearch"],
    )
    uid: str = Field(description="Grafana datasource UID, as assigned by Grafana.")
    grafana_uid: str = Field(description="Grafana UID.")


class GrafanaSourceAppData(BaseModel):
    """Application databag model for both sides of this interface."""

    datasources: Json[List[GrafanaDatasource]]


class DSExchangeAppData(cosl.interfaces.utils.DatabagModelV2, GrafanaSourceAppData):
    """App databag schema for both sides of this interface."""


class DatasourceDict(TypedDict):
    """Raw datasource information."""

    type: str
    uid: str
    grafana_uid: str


class EndpointValidationError(ValueError):
    """Raised if an endpoint name is invalid."""


def _validate_endpoints(
    charm: CharmBase, provider_endpoint: Optional[str], requirer_endpoint: Optional[str]
):
    meta = charm.meta
    for endpoint, source in (
        (provider_endpoint, meta.provides),
        (requirer_endpoint, meta.requires),
    ):
        if endpoint is None:
            continue
        if endpoint not in source:
            raise EndpointValidationError(f"endpoint {endpoint!r} not declared in charm metadata")
        interface_name = source[endpoint].interface_name
        if interface_name != DS_EXCHANGE_INTERFACE_NAME:
            raise EndpointValidationError(
                f"endpoint {endpoint} has unexpected interface {interface_name!r} "
                f"(should be {DS_EXCHANGE_INTERFACE_NAME})."
            )
    if not provider_endpoint and not requirer_endpoint:
        raise EndpointValidationError(
            "This charm should implement either a requirer or a provider (or both)"
            "endpoint for `grafana-datasource-exchange`."
        )


class DatasourceExchange:
    """``grafana_datasource_exchange`` interface endpoint wrapper (provider AND requirer)."""

    def __init__(
        self,
        charm: ops.CharmBase,
        *,
        provider_endpoint: Optional[str],
        requirer_endpoint: Optional[str],
    ):
        self._charm = charm
        _validate_endpoints(charm, provider_endpoint, requirer_endpoint)

        # gather all relations, provider or requirer
        all_relations = []
        if provider_endpoint:
            all_relations.extend(charm.model.relations.get(provider_endpoint, ()))
        if requirer_endpoint:
            all_relations.extend(charm.model.relations.get(requirer_endpoint, ()))

        # filter out some common unhappy relation states
        self._relations: List[ops.Relation] = [
            rel for rel in all_relations if (rel.app and rel.data)
        ]

    def publish(self, datasources: Iterable[DatasourceDict]):
        """Submit these datasources to all remotes.

        This operation is leader-only.
        """
        # sort by UID to prevent endless relation-changed cascades if this keeps flapping
        encoded_datasources = json.dumps(sorted(datasources, key=lambda raw_ds: raw_ds["uid"]))
        app_data = DSExchangeAppData(
            datasources=encoded_datasources  # type: ignore[reportCallIssue]
        )

        for relation in self._relations:
            app_data.dump(relation.data[self._charm.app])

    @property
    def received_datasources(self) -> Tuple[GrafanaDatasource, ...]:
        """Collect all datasources that the remotes have shared.

        This operation is leader-only.
        """
        datasources: List[GrafanaDatasource] = []

        for relation in self._relations:
            try:
                datasource = DSExchangeAppData.load(relation.data[relation.app])
            except DataValidationError:
                # load() already logs something in this case
                continue

            datasources.extend(datasource.datasources)
        return tuple(sorted(datasources, key=lambda ds: ds.uid))



from ops.framework import StoredDict, StoredList

from cosl.types import type_convert_stored


def test_converting_stored_types():
    # GIVEN empty structures
    assert type_convert_stored(StoredDict({}, under={})) == {}
    assert type_convert_stored(StoredList({}, under=[])) == []

    # GIVEN simple key-value pairs
    assert type_convert_stored(StoredDict({}, under={1: {}})) == {1: {}}
    assert type_convert_stored(StoredDict({}, under={1: []})) == {1: []}
    assert type_convert_stored(StoredDict({}, under={1: 2})) == {1: 2}

    # GIVEN nested structures and mixed types
    assert type_convert_stored(StoredList({}, under=[1, StoredList({}, under=[3, 4])])) == [
        1,
        [3, 4],
    ]
    assert type_convert_stored(StoredDict({}, under={1: StoredList({}, under=[3, 4])})) == {
        1: [3, 4]
    }
    assert type_convert_stored(StoredDict({}, under={1: StoredDict({}, under={2: 3})})) == {
        1: {2: 3}
    }
    assert type_convert_stored(StoredList({}, under=[1, StoredDict({}, under={2: 3})])) == [
        1,
        {2: 3},
    ]

    # GIVEN non-standard types
    assert type_convert_stored(StoredDict({}, under={1: "string", 2: 3.5})) == {
        1: "string",
        2: 3.5,
    }
    assert type_convert_stored(StoredList({}, under=[None, True, False])) == [None, True, False]

    # GIVEN deeply nested structures
    assert type_convert_stored(
        StoredDict(
            {},
            under={
                1: StoredList({}, under=[StoredDict({}, under={2: StoredList({}, under=[5])})])
            },
        )
    ) == {1: [{2: [5]}]}



# Copyright 2020 Canonical Ltd.
# See LICENSE file for licensing details.

import unittest

from cosl import *  # noqa: F403


class TestImport(unittest.TestCase):
    def test_all(self):
        # WHEN `import *` is used (see module-level import)
        # THEN all public symbols become available
        _ = CosTool  # noqa: F405
        _ = JujuTopology  # noqa: F405



"""Regretful reconciler charm utils."""

import inspect
import itertools
from typing import Any, Callable, Final, Iterable, Set, Type, TypeVar, Union, cast

import ops

_EventTyp = TypeVar("_EventTyp", bound=Type[ops.EventBase])
_EventBaseSubclassIterable = Iterable[_EventTyp]
_EventBaseSubclassSet = Set[_EventTyp]

# baseline obtained by:
# all_events = _get_inheritance_tree_leaves(ops.HookEvent)
# where:
# def _get_inheritance_tree_leaves(cl:Type):
#     return list(
#         itertools.chain(
#             *[
#                 ([subc] if (subc.__module__.startswith("ops.") and not subc.__subclasses__()) else
#                 _get_inheritance_tree_leaves(subc)) for subc in cl.__subclasses__()
#             ]
#         )
#     )
# at: 11/8/2025 (ops==2.17.1)
all_events: Final[Set[Type[ops.EventBase]]] = {
    ops.charm.PebbleCheckRecoveredEvent,
    ops.charm.PebbleCheckFailedEvent,
    ops.charm.ConfigChangedEvent,
    ops.charm.UpdateStatusEvent,
    ops.charm.PreSeriesUpgradeEvent,
    ops.charm.PostSeriesUpgradeEvent,
    ops.charm.LeaderElectedEvent,
    ops.charm.LeaderSettingsChangedEvent,
    # ops.charm.CollectMetricsEvent,           # deprecated
    ops.charm.RelationCreatedEvent,
    ops.charm.PebbleReadyEvent,
    ops.charm.RelationJoinedEvent,
    ops.charm.RelationChangedEvent,
    ops.charm.RelationDepartedEvent,
    ops.charm.RelationBrokenEvent,
    ops.charm.StorageAttachedEvent,
    ops.charm.StorageDetachingEvent,
    ops.charm.SecretChangedEvent,
    ops.charm.SecretRotateEvent,
    ops.charm.SecretRemoveEvent,
    ops.charm.SecretExpiredEvent,
    ops.charm.InstallEvent,
    ops.charm.StartEvent,
    ops.charm.RemoveEvent,
    ops.charm.StopEvent,
    ops.charm.UpgradeCharmEvent,
    ops.charm.PebbleCustomNoticeEvent,
}

# reconcilable_events_k8s is a list of all_events EXCEPT those listed below.
# Always include a comment with the reason for exclusions
reconcilable_events_k8s: Final[Set[Type[ops.EventBase]]] = all_events.difference(
    {
        ops.charm.PebbleCustomNoticeEvent,  # sometimes you want to handle the various notices differently
        ops.charm.RemoveEvent,  # usually pointless and sometimes harmful to reconcile towards an up state if you're shutting down
        ops.charm.UpgradeCharmEvent,  # this is your only chance to know you've been upgraded
    }
)

# reconcilable_events_machine is a list of all_events EXCEPT those listed below.
# Always include a comment with the reason for exclusions
reconcilable_events_machine: Final[Set[Type[ops.EventBase]]] = all_events.difference(
    {
        ops.charm.InstallEvent,  # (machine) charms may want to observe this
        ops.charm.RemoveEvent,  # usually pointless and sometimes harmful to reconcile towards an up state if you're shutting down
        ops.charm.StartEvent,  # (machine) charms may want to observe this
        ops.charm.StopEvent,  # usually pointless and sometimes harmful to reconcile towards an up state if you're shutting down
        ops.charm.UpgradeCharmEvent,  # this is your only chance to know you've been upgraded
    }
)


_CTR = itertools.count()


def observe_events(
    charm: ops.CharmBase,
    events: Iterable[_EventTyp],
    handler: Union[Callable[[Any], None], Callable[[], None]],
):
    """Observe all events that are subtypes of a given list using the provided handler.

    Usage:
    >>> class MyCharm(ops.CharmBase):
    ...    def __init__(self, ...):
    ...        super().__init__(...)
    ...        observe_events(self, all_events, self.reconcile)
    ...
    ...     def reconcile(self):
    ...         pass

    Or:
    >>> class MyK8sCharm(ops.CharmBase):
    ...    def __init__(self, ...):
    ...        super().__init__(...)
    ...        observe_events(self, reconcilable_events_k8s, self._on_any_event)
    ...
    ...    def _on_any_event(self, _):
    ...        self.reconcile()
    ...
    ...     def reconcile(self):
    ...         pass

    Or:
    >>> class MyFineGrainedCharm(ops.CharmBase):
    ...    def __init__(self, ...):
    ...        super().__init__(...)
    ...        observe_events(self, {ops.StartEvent, ops.StopEvent}, self._on_group1)
    ...        observe_events(self, {ops.RelationEvent, ops.SecretEvent, ops.framework.LifecycleEvent}, self._on_group2,)
    ...        # ... add more groups as needed
    ...
    ...    def _on_group1(self):
    ...        pass
    ...
    ...     def _on_group2(self):
    ...        pass

    Or even! (but you're a bad person if you do this)
    >>> class MyCharm(ops.CharmBase):
    ...    def __init__(self, ...):
    ...        super().__init__(...)
    ...        observe_events(self, {ops.RelationEvent}, lambda: print("I am running a relation event"))
    """
    # ops types it with Any!
    evthandler: Callable[[Any], None]
    if not inspect.signature(handler).parameters:
        # handler provided is a function not part of an ops.Object
        class _Observer(ops.Object):
            _key = f"_observer_proxy_{next(_CTR)}"

            def __init__(self):
                super().__init__(charm, key=self._key)
                # attach ref to something solid to prevent inadvertently GC'ing this thang
                setattr(charm.framework, self._key, self)

            def evt_handler(self, _: ops.EventBase) -> None:
                handler()  # type: ignore

        evthandler = _Observer().evt_handler
    else:
        # handler provided is a method of an ops.Object
        evthandler = cast(Callable[[Any], None], handler)

    for bound_evt in charm.on.events().values():
        if any(issubclass(bound_evt.event_type, include_type) for include_type in events):
            charm.framework.observe(bound_evt, evthandler)



# Copyright 2026 Canonical Ltd.
# See LICENSE file for licensing details.
"""Test the round-trip encoding/decoding of the LZMABase64 class."""

import json

import pytest

from cosl import LZMABase64


def test_decompress_invalid_data():
    # Should raise an error for corrupted input
    with pytest.raises(Exception):
        LZMABase64.decompress("not-a-valid-base64")


@pytest.mark.parametrize("non_str", [123, 45.6, [1, 2, 3], {"a": 1}])
def test_compress_non_string(non_str):
    with pytest.raises(Exception):
        LZMABase64.compress(non_str)


def test_round_trip_dict():
    d = {
        "some": "dict",
        "with": "keys",
        "even": [{"nested": "types", "and_integers": [42, 42]}],
    }
    assert d == json.loads(LZMABase64.decompress(LZMABase64.compress(json.dumps(d))))


@pytest.mark.parametrize(
    "input",
    [
        "simple string",
        "string with emojis: 😀🚀🌟",
        "",  # empty string
        " " * 100,  # whitespace only
        "!@#$%^&*()_+-=[]{}|;':,./<>?",  # special characters
        "\n\t\r",  # only escape sequences
        "a" * 10000,  # very large string
        bytes([0, 255, 127, 10, 20, 30]).hex(),  # binary data encoded as string
    ],
)
def test_round_trip(input):
    assert input == LZMABase64.decompress(LZMABase64.compress(input))



import json

import pytest

from cosl.interfaces.utils import DatabagModel, DataValidationError


class MyDataModel(DatabagModel):
    foo: int
    bar: str


def test_dump():
    # given a databag model subclass
    dm = MyDataModel(foo=1, bar="baz")

    # when you dump it
    out = dm.dump()

    # then you obtain a json-dict representation, flat
    assert out == {"foo": json.dumps(1), "bar": json.dumps("baz")}


def test_dump_to_mapping():
    # given a databag model subclass
    dm = MyDataModel(foo=1, bar="baz")

    # when you dump it to an arbitrary mapping
    databag = {"spurious": "data"}
    out = dm.dump(databag=databag)

    # then you obtain it back, populated
    assert databag is out
    assert databag == {"foo": json.dumps(1), "bar": json.dumps("baz")}


def test_dump_to_mapping_append():
    # given a databag model subclass
    dm = MyDataModel(foo=1, bar="baz")

    # when you dump it to an arbitrary mapping, without clearing it first
    databag = {"spurious": "data"}
    out = dm.dump(databag=databag, clear=False)

    # then you obtain it back, populated and with the existing data still there
    assert databag is out
    assert databag == {"foo": json.dumps(1), "bar": json.dumps("baz"), "spurious": "data"}


@pytest.mark.parametrize(
    "example",
    (
        {"foo": 1, "bar": "baz"},
        {"foo": 10, "bar": "lol"},
        {"foo": 11, "bar": "hola hola", "extra": "field"},
    ),
)
def test_load(example):
    # given a databag model subclass
    # when you instantiate it from an existing mapping
    dm = MyDataModel.load({key: json.dumps(value) for key, value in example.items()})
    # then you obtain a validated instance
    assert isinstance(dm, MyDataModel)
    assert dm.foo == example["foo"]


@pytest.mark.parametrize(
    "example",
    (
        {"foo": 1, "bar": 1},
        {"foo": [1, 2], "bar": "lol"},
        {"bar": "hola"},
    ),
)
def test_load_invalid_raises(example):
    # given a databag model subclass
    # when you instantiate it from an existing mapping with bad data
    # then you get an error
    with pytest.raises(DataValidationError):
        MyDataModel.load({key: json.dumps(value) for key, value in example.items()})



# Copyright 2026 Canonical Ltd.
# See LICENSE file for licensing details.

import json
import unittest

from cosl import DashboardPath40UID, GrafanaDashboard


class TestRoundTripEncDec(unittest.TestCase):
    """Tests the round-trip encoding/decoding of the GrafanaDashboard class."""

    def test_round_trip(self):
        d = {
            "some": "dict",
            "with": "keys",
            "even": [{"nested": "types", "and_integers": [42, 42]}],
        }
        self.assertDictEqual(d, GrafanaDashboard._serialize(json.dumps(d))._deserialize())


class TestGenerateUID(unittest.TestCase):
    """Spec for the UID generation logic."""

    def test_generate_no_arguments_raises_error(self):
        """Test that generate raises ValueError when no arguments are provided."""
        with self.assertRaises(ValueError) as context:
            DashboardPath40UID.generate()
        self.assertIn("At least one string argument is required", str(context.exception))

    def test_generate_basic_functionality(self):
        """Test basic UID generation with common scenarios."""
        # Test with single argument
        uid1 = DashboardPath40UID.generate("my-charm")
        self.assertEqual(40, len(uid1))
        self.assertTrue(DashboardPath40UID.is_valid(uid1))

        # Test with multiple arguments
        uid2 = DashboardPath40UID.generate("my-charm", "dashboard.json", "v2", "production")
        self.assertEqual(40, len(uid2))
        self.assertTrue(DashboardPath40UID.is_valid(uid2))

        # Test backward compatibility with original signature
        uid3 = DashboardPath40UID.generate("some-charm", "dashboard.json")
        self.assertEqual(40, len(uid3))
        self.assertTrue(DashboardPath40UID.is_valid(uid3))

    def test_generate_deterministic_behavior(self):
        """Test that the same arguments produce the same UID (deterministic)."""
        test_cases = [
            ("single_arg",),
            ("my-charm", "dashboard.json"),
            ("my-charm", "dashboard.json", "v2", "production"),
        ]

        for args in test_cases:
            uid1 = DashboardPath40UID.generate(*args)
            uid2 = DashboardPath40UID.generate(*args)
            self.assertEqual(uid1, uid2, f"UIDs should be identical for args: {args}")

    def test_generate_uniqueness_across_combinations(self):
        """Test that different argument combinations produce unique UIDs."""
        combinations = [
            ("some-charm", "dashboard1.json"),
            ("some-charm", "dashboard2.json"),
            ("some-charm", "tmp/dashboard1.json"),
            ("diff-charm", "dashboard1.json"),
            ("arg1",),
            ("arg1", "arg2"),
            ("arg1", "arg2", "arg3"),
            ("arg2", "arg1"),  # Different order
        ]

        uids = [DashboardPath40UID.generate(*args) for args in combinations]

        # All UIDs should be unique
        for i in range(len(uids)):
            for j in range(i + 1, len(uids)):
                self.assertNotEqual(
                    uids[i],
                    uids[j],
                    f"UIDs should be different: {combinations[i]} vs {combinations[j]}",
                )

    def test_is_valid_edge_cases(self):
        """Test validity check with edge cases."""
        # Invalid cases
        self.assertFalse(DashboardPath40UID.is_valid("1234"))
        self.assertFalse(DashboardPath40UID.is_valid("short non-hex string"))
        self.assertFalse(DashboardPath40UID.is_valid("non-hex string, crafted to be 40 chars!!"))
        self.assertFalse(DashboardPath40UID.is_valid(""))
        self.assertFalse(DashboardPath40UID.is_valid(None))  # type: ignore
        self.assertFalse(DashboardPath40UID.is_valid(False))  # type: ignore

        # Valid cases
        self.assertTrue(DashboardPath40UID.is_valid("0" * 40))
        self.assertTrue(
            DashboardPath40UID.is_valid("a1b2c3d4e5f6789012345678901234567890abcd")
        )  # 40 chars
        # Generated UIDs should always be valid (covered by other tests but explicitly stated here)
        sample_uid = DashboardPath40UID.generate("some-charm", "dashboard.json")
        self.assertTrue(DashboardPath40UID.is_valid(sample_uid))



# Copyright 2026 Canonical Ltd.
# See LICENSE file for licensing details.
"""Utility functions and classes."""

import base64
import lzma
from typing import Union


class LZMABase64:
    """A helper class for LZMA-compressed-base64-encoded strings.

    This is useful for transferring over juju relation data, which can only have keys of type string.
    """

    @classmethod
    def compress(cls, raw_json: Union[str, bytes]) -> str:
        """LZMA-compress and base64-encode into a string."""
        if not isinstance(raw_json, bytes):
            raw_json = raw_json.encode("utf-8")
        return base64.b64encode(lzma.compress(raw_json)).decode("utf-8")

    @classmethod
    def decompress(cls, compressed: str) -> str:
        """Decompress from base64-encoded-lzma-compressed string."""
        return lzma.decompress(base64.b64decode(compressed.encode("utf-8"))).decode()
